/*
 * Copyright (c) 2016 Chris Johns <chrisj@rtems.org>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#define TARGET_DEBUG 1

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <errno.h>
#include <inttypes.h>
#include <stdlib.h>

#include <rtems.h>
#include <rtems/score/threadimpl.h>

#include <libcpu/arm-cp15.h>
#include <bsp/linker-symbols.h>

#include "rtems-dbserver-target.h"
#include "rtems-dbserver-threads.h"

#if TARGET_DEBUG
#include <rtems/bspIo.h>
#endif

/*
 * Software breakpoint block size.
 */
#define RTEMS_DBSERVER_SWBREAK_NUM 64

/*
 * Hardware breakpoints. Limited by hardware
 */
#define RTEMS_DBSERVER_HWBREAK_NUM 4

/*
 * Number of registers.
 */
#define RTEMS_DBSERVER_NUMREGS 26

/*
 * Number of bytes per register.
 */
#define RTEMS_DBSERVER_REGBYTES 4

/*
 * Number of bytes of registers.
 */
#define RTEMS_DBSERVER_NUMREGBYTES \
  (RTEMS_DBSERVER_NUMREGS * RTEMS_DBSERVER_REGBYTES)

/*
 * Debugger registers layout.
 */
#define REG_R0    0
#define REG_R1    1
#define REG_R2    2
#define REG_R3    3
#define REG_R4    4
#define REG_R5    5
#define REG_R6    6
#define REG_R7    7
#define REG_R8    8
#define REG_R9    9
#define REG_R10   10
#define REG_R11   11
#define REG_R12   12
#define REG_SP    13
#define REG_LR    14
#define REG_PC    15
#define REG_CPSR  25

/**
 * Currently default to the debug macrocell being present in the device. The
 * SWI code is not tested and may go away.
 */
#define ARM_USE_DEBUG_MACROCELL 1

/**
 * The various status registers.
 */
#if defined(ARM_MULTILIB_ARCH_V4)
 #define FRAME_SR frame->register_cpsr
#elif defined(ARM_MULTILIB_ARCH_V7M)
 #define FRAME_SR frame->register_xpsr
#else
 #error ARM architecture is not supported.
#endif

/**
 * Code to switch between ARM and Thumb mode.
 */
#ifdef __thumb__
 #define SWITCH_ARM_TO_THUMB ".arm\nadd r1,pc,#1\nbx r1\n"
 #define SWITCH_THUMB_TO_ARM ".thumb\n.align 2\nbx pc\n.arm\n"
#else
 #define SWITCH_ARM_TO_THUMB
 #define SWITCH_THUMB_TO_ARM
#endif /* __thumb__ */

#if ARM_USE_SWI
 /**
  * The breakpoint is Supervisor Call (SVC) or what use to be called 'SWI'.
  */
 #ifdef __thumb__
  static const uint8_t breakpoint[2] = { 0x55, 0xdf };
 #else
  static const uint8_t breakpoint[4] = { 0x00, 0xef, 0x55, 0x00 };
 #endif
#endif /* ARM_USE_SWI */

#if ARM_USE_DEBUG_MACROCELL
 /**
  * The breakpoint is Supervisor Call (SVC) or what use to be called 'SWI'.
  */
 #ifdef __thumb__
  static const uint8_t breakpoint[2] = { 0x55, 0xbe };
 #else
  static const uint8_t breakpoint[4] = { 0x75, 0xe0a, 0x20, 0xe1 };
 #endif
#endif

/**
 * Target lock.
 */
RTEMS_INTERRUPT_LOCK_DEFINE(static, target_lock, "target_lock")

/**
 * The init value for the text section.
 */
static uint32_t text_section_flags;

/**
 * Is a session active?
 */
static bool debug_session_active;

#if TARGET_DEBUG
static void target_printk(const char* format, ...) RTEMS_PRINTFLIKE(1, 2);
static void
target_printk(const char* format, ...)
{
  va_list ap;
  va_start(ap, format);
  vprintk(format, ap);
  va_end(ap);
}
#else
#define target_printk(_fmt, ...)
#endif

#if ARM_USE_DEBUG_MACROCELL
/*
 * Use the ARM debug hardware.
 */
static int debug_version;
static int debug_revision;;
static int hw_breakpoints;
static int hw_watchpoints;
static uint32_t
arm_dbgdidr_read(void)
{
  ARM_SWITCH_REGISTERS;
  uint32_t val;
  __asm__ volatile (
    ARM_SWITCH_TO_ARM
    "mrc p14, 0, %[val], c0, c0, 0\n"
    ARM_SWITCH_BACK
    : [val] "=&r" (val) ARM_SWITCH_ADDITIONAL_OUTPUT
  );
  return val;
}

static void
arm_debug_macrocell(rtems_dbserver_target* target)
{
  #define ID_VALUE(_i, _h, _l) ((_i >> _l) & ((1 << ((_h - _l) + 1)) -1))
  uint32_t id = arm_dbgdidr_read();
  debug_version = ID_VALUE(id, 19, 16);
  debug_revision = ID_VALUE(id, 3, 0);
  hw_breakpoints = ID_VALUE(id, 27, 24);
  hw_watchpoints = ID_VALUE(id, 31, 28);
  rtems_dbserver_printf("rtems-db: arm debug: (v%d.%d) hbreak:%d hwatch:%d\n",
                        debug_version, debug_revision,
                        hw_breakpoints, hw_watchpoints);
}
#endif /* ARM_USE_DEBUG_MACROCELL */

int
rtems_dbserver_target_configure(rtems_dbserver_target* target)
{
  target->capabilities = (RTEMS_DBSERVER_TARGET_CAP_SWBREAK);
  target->reg_num = RTEMS_DBSERVER_NUMREGS;
  target->reg_size = sizeof(uint32_t);
  target->breakpoint = &breakpoint[0];
  target->breakpoint_size = sizeof(breakpoint);
  arm_debug_macrocell(target);
  return 0;
}

static void
target_exception(CPU_Exception_frame* frame)
{
  target_printk("[} frame = %08lx sig=%d vector=%x\n",
                (uint32_t) frame,
                rtems_dbserver_target_exception_to_signal(frame),
                frame->vector);
  target_printk("[}  R0 = %08" PRIx32 "  R1 = %08" PRIx32             \
                "  R2 = %08" PRIx32 "  R3 = %08" PRIx32 "\n",
                frame->register_r0, frame->register_r1,
                frame->register_r2, frame->register_r3);
  target_printk("[}  R4 = %08" PRIx32 "  R5 = %08" PRIx32             \
                "  R6 = %08" PRIx32 "  R7 = %08" PRIx32 "\n",
                frame->register_r4, frame->register_r5,
                frame->register_r6, frame->register_r7);
  target_printk("[}  R8 = %08" PRIx32 "  R9 = %08" PRIx32             \
                " R10 = %08" PRIx32 " R11 = %08" PRIx32 "\n",
                frame->register_r8, frame->register_r9,
                frame->register_r10, frame->register_r11);
  target_printk("[} R12 = %08" PRIx32 "\n", frame->register_r12);
  target_printk("[}  SP = %08" PRIx32 "  LR = %08" PRIxPTR           \
                "  PC = %08" PRIxPTR " CPSR = %08" PRIx32 "\n",
                frame->register_sp, (intptr_t) frame->register_lr,
                (intptr_t) frame->register_pc, FRAME_SR);

  if (!debug_session_active)
    _ARM_Exception_default(frame);

  switch (rtems_dbserver_target_exception(frame)) {
  case rtems_dbserver_target_exc_consumed:
  default:
    break;
  case rtems_dbserver_target_exc_step:
    break;
  case rtems_dbserver_target_exc_cascade:
    target_printk("rtems-db: unhandled exception: cascading\n");
    _ARM_Exception_default(frame);
    break;
  }

  target_printk("[} resuming frame = %08lx\n", (uint32_t) frame);
}

/**
 * Create an exception stack frame.
 *
 * We will have switched from user or svc to an exception mode. Create an
 * exception frame on the exception's stack and then copy it to the thread's
 * stack. Switch back to the thread's context and mode to handle the exception
 * to avoid any stack checks thinking the stack is blown.
 */
#define EXCEPTION_FRAME_CREATE(_frame, _pc_offset)                   \
  __asm__ volatile(                                                  \
    "sub  sp, %[frame_size]\n"                /* alloc the frame */  \
    "stm  sp, {r0-r12}\n"                        /* store r0-r12 */  \
    "add  r0, sp, %[r0_r12_size]\n"   /* get the sp in the frame */  \
    "mrs  r1, spsr\n"                        /* get the saved sr */  \
    "mov  r6, r1\n"                        /* stash it for later */  \
    "bic  r1, r1, %[psr_t]\n"     /* clear thumb mode, not sure? */  \
    "orr  r1, r1, %[psr_i]\n"                       /* mask irqs */  \
    "mrs  r2, cpsr\n"                      /* get the current sr */  \
    "msr  cpsr, r1\n"                     /* switch to user mode */  \
    "mov  r3, sp\n"                     /* get the stack pointer */  \
    "mov  r4, lr\n"                          /* get the link reg */  \
    "msr  cpsr, r2\n"                       /* back to this mode */  \
    "sub  r5, lr, %[pc_offset]\n"         /* get the adjusted pc */  \
    "stm  r0, {r3-r6}\n"                  /* save into the frame */  \
    "sub  r4, r3, %[frame_size]\n"        /* destination address */  \
    "mov  r6, r4\n"                            /* save the frame */  \
    "sub  r4, #1\n"                      /* one before the start */  \
    "add  r3, #1\n"                          /* one past the end */  \
    "sub  r5, sp, #1\n"                        /* source address */  \
    "1:\n"                                                           \
    "ldrb r0, [r5, #1]!\n"                         /* get a byte */  \
    "strb r0, [r4, #1]!\n"                       /* put the byte */  \
    "cmp  r3, r4\n"                                  /* the end? */  \
    "bne  1b\n"                                                      \
    "add  sp, %[frame_size]\n"                 /* free the frame */  \
    "mrs  r1, spsr\n"               /* get the thread's saved sr */  \
    "orr  r2, r1, %[psr_i]\n"                       /* mask irqs */  \
    "msr  cpsr, r2\n"     /* switch back to the thread's context */  \
    "sub  sp, %[frame_size]\n"      /* alloc the frame in thread */  \
    "msr  cpsr, r1\n"                    /* restore the irq mask */  \
    "mov  %[o_frame], sp\n"                    /* save the frame */  \
    : [o_frame] "=r" (_frame)                                        \
    : [psr_t] "i" (ARM_PSR_T),                                       \
      [psr_i] "i" (ARM_PSR_I),                                       \
      [pc_offset] "i" (_pc_offset),                                  \
      [r0_r12_size] "i" (13 * sizeof(uint32_t)),                     \
      [frame_size] "i" (sizeof(CPU_Exception_frame))                 \
    : "r0", "r1", "r2", "r3", "r4", "r5", "r6", "memory")

#define EXCEPTION_FRAME_DESTROY(_pc_offset)                          \
  __asm__ volatile(                                                  \
    "ldm    sp, {r0-r12}\n"                                          \
    "add    sp, %[frame_size]\n"                                     \
    "subs   pc, lr, #4\n"                                            \
    :                                                                \
    : [frame_size] "i" (sizeof(CPU_Exception_frame)),                \
      [pc_offset] "i" (_pc_offset))


static void __attribute__((naked))
target_exception_undefined_instruction(void)
{
  CPU_Exception_frame* frame;
  EXCEPTION_FRAME_CREATE(frame, 4);
  frame->vector = 1;
  target_exception(frame);
  EXCEPTION_FRAME_DESTROY(4);
}

#if ARM_USE_SWI
static void __attribute__((naked))
target_exception_supervisor_call(void)
{
  CPU_Exception_frame* frame;
  EXCEPTION_FRAME_CREATE(frame, 4);
  frame->vector = 2;
  target_exception(frame);
  EXCEPTION_FRAME_DESTROY(4);
}
#endif /* ARM_USE_SWI */

static void __attribute__((naked))
target_exception_prefetch_abort(void)
{
  CPU_Exception_frame* frame;
  EXCEPTION_FRAME_CREATE(frame, 4);
  frame->vector = 3;
  target_exception(frame);
  EXCEPTION_FRAME_DESTROY(4);
}

static void __attribute__((naked))
target_exception_data_abort(void)
{
  CPU_Exception_frame* frame;
  EXCEPTION_FRAME_CREATE(frame, 4);
  frame->vector = 4;
  target_exception(frame);
  EXCEPTION_FRAME_DESTROY(4);
}

int
rtems_dbserver_target_enable(void)
{
  rtems_interrupt_lock_context lock_context;
  void*                        text_begin;
  void*                        text_end;
  debug_session_active = true;
  text_begin = &bsp_section_text_begin[0];
  text_end = &bsp_section_text_end[0];
  rtems_interrupt_lock_acquire(&target_lock, &lock_context);
  text_section_flags =
    arm_cp15_set_translation_table_entries(text_begin,
                                           text_end,
                                           ARMV7_MMU_DATA_READ_WRITE_CACHED);
  arm_cp15_set_exception_handler(ARM_EXCEPTION_UNDEF,
                                 target_exception_undefined_instruction);
#if ARM_USE_SWI
  arm_cp15_set_exception_handler(ARM_EXCEPTION_SWI,
                                 target_exception_supervisor_call);
#endif
  arm_cp15_set_exception_handler(ARM_EXCEPTION_PREF_ABORT,
                                 target_exception_prefetch_abort);
  arm_cp15_set_exception_handler(ARM_EXCEPTION_DATA_ABORT,
                                 target_exception_data_abort);
  rtems_interrupt_lock_release(&target_lock, &lock_context);
  return 0;
}

int
rtems_dbserver_target_disable(void)
{
  rtems_interrupt_lock_context lock_context;
  void*                        text_begin;
  void*                        text_end;
  rtems_interrupt_lock_acquire(&target_lock, &lock_context);
  debug_session_active = false;
  text_begin = &bsp_section_text_begin[0];
  text_end = &bsp_section_text_end[0];
  arm_cp15_set_translation_table_entries(text_begin,
                                         text_end,
                                         text_section_flags);
  rtems_interrupt_lock_release(&target_lock, &lock_context);
  return 0;
}

int
rtems_dbserver_target_read_regs(rtems_dbserver_thread* thread)
{
  if (!rtems_dbserver_thread_flag(thread,
                                  RTEMS_DBSERVER_THREAD_FLAG_REG_VALID)) {
    static const uint32_t good_address = (uint32_t) &good_address;
    uint32_t*             regs = &thread->registers[0];
    int                   i;

    for (i = 0; i < RTEMS_DBSERVER_NUMREGS; ++i)
      regs[i] = (uint32_t) &good_address;

    if (thread->frame) {
      CPU_Exception_frame* frame = thread->frame;
      regs[REG_R0]   = frame->register_r0;
      regs[REG_R1]   = frame->register_r1;
      regs[REG_R2]   = frame->register_r2;
      regs[REG_R3]   = frame->register_r3;
      regs[REG_R4]   = frame->register_r4;
      regs[REG_R5]   = frame->register_r5;
      regs[REG_R6]   = frame->register_r6;
      regs[REG_R7]   = frame->register_r7;
      regs[REG_R8]   = frame->register_r8;
      regs[REG_R9]   = frame->register_r9;
      regs[REG_R10]  = frame->register_r10;
      regs[REG_R11]  = frame->register_r11;
      regs[REG_R12]  = frame->register_r12;
      regs[REG_SP]   = frame->register_sp;
      regs[REG_LR]   = (uint32_t) frame->register_lr;
      regs[REG_PC]   = (uint32_t) frame->register_pc;
      regs[REG_CPSR] = FRAME_SR;
      /*
       * Get the signal from the frame.
       */
      thread->signal = rtems_dbserver_target_exception_to_signal(frame);
    }
    else {
#if defined(ARM_MULTILIB_ARCH_V4)
      regs[REG_R4]  = thread->tcb->Registers.register_r4;
      regs[REG_R5]  = thread->tcb->Registers.register_r5;
      regs[REG_R6]  = thread->tcb->Registers.register_r6;
      regs[REG_R7]  = thread->tcb->Registers.register_r7;
      regs[REG_R8]  = thread->tcb->Registers.register_r8;
      regs[REG_R9]  = thread->tcb->Registers.register_r9;
      regs[REG_R10] = thread->tcb->Registers.register_r10;
      regs[REG_R11] = thread->tcb->Registers.register_fp;
      regs[REG_LR]  = (intptr_t) thread->tcb->Registers.register_lr;
      regs[REG_PC]  = (intptr_t) thread->tcb->Registers.register_lr;
      regs[REG_SP]  = (intptr_t) thread->tcb->Registers.register_sp;
#elif defined(ARM_MULTILIB_ARCH_V7M)
      regs[REG_R4]  = thread->tcb->Registers.register_r4;
      regs[REG_R5]  = thread->tcb->Registers.register_r5;
      regs[REG_R6]  = thread->tcb->Registers.register_r6;
      regs[REG_R7]  = thread->tcb->Registers.register_r7;
      regs[REG_R8]  = thread->tcb->Registers.register_r8;
      regs[REG_R9]  = thread->tcb->Registers.register_r9;
      regs[REG_R10] = thread->tcb->Registers.register_r10;
      regs[REG_R11] = thread->tcb->Registers.register_r11;
      regs[REG_LR]  = (intptr_t) thread->tcb->Registers.register_lr;
      regs[REG_PC]  = (intptr_t) thread->tcb->Registers.register_lr;
      regs[REG_SP]  = (intptr_t) thread->tcb->Registers.register_sp;
#endif
      /*
       * Blocked threads have no signal.
       */
      thread->signal = 0;
    }

    thread->flags |= RTEMS_DBSERVER_THREAD_FLAG_REG_VALID;
    thread->flags &= ~RTEMS_DBSERVER_THREAD_FLAG_REG_DIRTY;
  }

  return 0;
}

int
rtems_dbserver_target_write_regs(rtems_dbserver_thread* thread)
{
  if (rtems_dbserver_thread_flag(thread,
                                 RTEMS_DBSERVER_THREAD_FLAG_REG_DIRTY)) {
    uint32_t* regs = &thread->registers[0];

    /*
     * Only write to debugger controlled threads. Do not touch the registers
     * for threads blocked in the context switcher.
     */
    if (rtems_dbserver_thread_flag(thread,
                                   RTEMS_DBSERVER_THREAD_FLAG_DEBUGGING)) {
      CPU_Exception_frame* frame = thread->frame;
      frame->register_r0  = regs[REG_R0];
      frame->register_r1  = regs[REG_R1];
      frame->register_r2  = regs[REG_R2];
      frame->register_r3  = regs[REG_R3];
      frame->register_r4  = regs[REG_R4];
      frame->register_r5  = regs[REG_R5];
      frame->register_r6  = regs[REG_R6];
      frame->register_r7  = regs[REG_R7];
      frame->register_r8  = regs[REG_R8];
      frame->register_r9  = regs[REG_R9];
      frame->register_r10 = regs[REG_R10];
      frame->register_r11 = regs[REG_R11];
      frame->register_r12 = regs[REG_R12];
      frame->register_sp  = regs[REG_SP];
      frame->register_lr  = (void*) regs[REG_LR];
      frame->register_pc  = (void*) regs[REG_PC];
      FRAME_SR            = regs[REG_CPSR];
    }
    thread->flags &= ~RTEMS_DBSERVER_THREAD_FLAG_REG_DIRTY;
  }
  return 0;
}

DB_UINT
rtems_dbserver_target_reg_pc(rtems_dbserver_thread* thread)
{
  int r;
  r = rtems_dbserver_target_read_regs(thread);
  if (r >= 0) {
    uint32_t* regs = &thread->registers[0];
    return regs[REG_PC];
  }
  return 0;
}

DB_UINT
rtems_dbserver_target_frame_pc(CPU_Exception_frame* frame)
{
  return (DB_UINT) frame->register_pc;
}

DB_UINT
rtems_dbserver_target_reg_sp(rtems_dbserver_thread* thread)
{
  int r;
  r = rtems_dbserver_target_read_regs(thread);
  if (r >= 0) {
    uint32_t* regs = &thread->registers[0];
    return regs[REG_SP];
  }
  return 0;
}

DB_UINT
rtems_dbserver_target_tcb_sp(rtems_dbserver_thread* thread)
{
  return (DB_UINT) thread->tcb->Registers.register_sp;
}

int
rtems_dbserver_target_thread_stepping(rtems_dbserver_thread* thread)
{
  if (rtems_dbserver_thread_flag(thread,
                                 (RTEMS_DBSERVER_THREAD_FLAG_STEP |
                                  RTEMS_DBSERVER_THREAD_FLAG_STEPPING))) {
    //CPU_Exception_frame* frame = thread->frame;
  }
  return 0;
}

int
rtems_dbserver_target_exception_to_signal(CPU_Exception_frame* frame)
{
  int sig = RTEMS_DBSERVER_SIGNAL_HUP;
  switch (frame->vector) {
  case ARM_EXCEPTION_RESET:
  case ARM_EXCEPTION_SWI:
    sig = RTEMS_DBSERVER_SIGNAL_TRAP;
    break;
  case ARM_EXCEPTION_UNDEF:
    sig = RTEMS_DBSERVER_SIGNAL_ILL;
    break;
  case ARM_EXCEPTION_FIQ:
    sig = RTEMS_DBSERVER_SIGNAL_FPE;
    break;
  case ARM_EXCEPTION_PREF_ABORT:
  case ARM_EXCEPTION_DATA_ABORT:
    sig = RTEMS_DBSERVER_SIGNAL_SEGV;
    break;
  case ARM_EXCEPTION_RESERVED:
  case ARM_EXCEPTION_IRQ:
    sig = RTEMS_DBSERVER_SIGNAL_BUS;
    break;
  default:
    break;
  }
  return sig;
}

int
rtems_dbserver_target_hwbreak_control(rtems_dbserver_target_watchpoint wp,
                                      bool                             insert,
                                      DB_UINT                          addr,
                                      DB_UINT                          kind)
{
  /*
   * To do.
   */
  return 0;
}

int
rtems_dbserver_target_cache_sync(rtems_dbserver_target_swbreak* swbreak)
{
  /*
   * Flush the data cache and invalidate the instruction cache.
   */
  rtems_cache_flush_multiple_data_lines(swbreak->address,
                                        sizeof(breakpoint));
  rtems_cache_invalidate_multiple_instruction_lines(swbreak->address,
                                                    sizeof(breakpoint));
  return 0;
}
